/*
 * nffs.h
 *
 *  Created on: 2021/4/11
 *      Author: ueyudiud
 */

#ifndef NFFS_H_
#define NFFS_H_

#include <stdint.h>

typedef unsigned char n_bool;

typedef int8_t n_i8;
typedef int16_t n_i16;
typedef int32_t n_i32;
typedef int64_t n_i64;

typedef uint8_t n_u8;
typedef uint16_t n_u16;
typedef uint32_t n_u32;
typedef uint64_t n_u64;

typedef char const* n_cstr;
typedef n_i32 n_code;
typedef void* n_handle;

#define addr2bid(d,a)		((a) / (d)->_param._bytes_per_block * (d)->->_param._bytes_per_block)

/**
 * @defgroup 端序操作
 * @{
 */
#define ne2u16(v)	(v)
#define ne2u32(v)	(v)
#define nu2e16(v)	(v)
#define nu2e32(v)	(v)
#define nc2e16(v)   ((n_u16) { v })
#define nc2e32(v)   ((n_u32) { v })
/** @} */

typedef struct nffsDeviceParam	n_dev_param;
typedef struct nffsDevVtable	n_dev_vtable;
typedef struct nffsDevice		n_dev;
typedef struct nffsBlock 		n_block;
typedef struct nffsINode 		n_inode;
typedef struct nffsDirent       n_dirent;
typedef struct nffsRawNodeRef	n_raw_node_ref;

/* NFFS 外部表示形式接口 */
typedef struct nffsFile         n_file;

struct nffsDeviceParam {
	n_cstr				_name;

	n_u32				_bytes_per_chunk;
	n_u32				_chunk_per_block;
	n_u32               _block_per_total;
	n_u32				_bytes_per_block;
	n_u32               _bytes_per_total;
	n_u32				_min_block_index;
	n_u32				_max_block_index;
};

#include "nffs_list.h"

/* NFFS low level interface */
#define memcpy(a, b, c)     __builtin_memcpy(a, b, c)
#define memset(a, b, c)     __builtin_memset(a, b, c)

#include <stdlib.h>
#define nliAllocRaw(sz)     malloc(sz)
#define nliAllocZero(sz)    calloc(1, sz)
#define nliAllocPage(sz)    malloc(sz)
#define nliFree(b,sz)       free(b)
#define nliFreePage(b)      free(b)

#define nliAllocDevice()    ((n_dev*) nliAllocZero(sizeof(n_dev)))
#define nliFreeDevice(dev)  nliFree(dev, sizeof(n_dev))

#define nliCrc32(d,l)       (((void) (d)), ((void) (l)), 0) /* TODO 功能相当于CRC32 */

/* NFFS miscellaneous function */

#define NMF_FALSE           0
#define NMF_TRUE            1

#define NMF_SUCC			0
#define NMF_FAIL(c)         (-(c))

#define NMF_NOMEM			1

#define nmfIsSuccess(c)		((c) >= 0)
#define nmfIsError(c)		((c) < 0)
#define nmfProtect(e)		do { n_code _c = (e); if (nmfIsError(_c)) return _c; } while (0)
#define nmfGetError(c)		(-(c))

#define nmfLikely(e)		(e)
#define nmfUnlikely(e)		(e)
#define nmfMax(a,b)         ((a) >= (b) ? (a) : (b))
#define nmfMin(a,b)         ((a) <= (b) ? (a) : (b))

#define NMF_ENTITY_DEV		0
#define NMF_ENTITY_KER		1
#define NMF_ENTITY_SELF		2
#define NMF_ENTITY_USER		3

#define NMF_DBG_LEVEL_TRACE	0
#define NMF_DBG_LEVEL_INFO	1
#define NMF_DBG_LEVEL_WARN	2
#define NMF_DBG_LEVEL_ERROR	3

#define nmfLog(e,l,f,a...)	((void) (e), (void) (l), nffsDummyLog(f,##a))

#define nmfTrap()           __builtin_trap()
#define nmfAssume(e,f,a...)	do if (!nmfLikely(e)) { nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_ERROR, f,##a); nmfTrap(); } while(0)

__attribute__((format(printf, 1, 2)))
extern inline void nffsDummyLog(n_cstr fmt, ...) { /* TODO log messages optionally? */ }

struct nffsDevVtable {
	n_code (*_write)(n_dev* dev, n_u32 addr, n_u32 len, n_u32* rlen, void const* data);
	n_code (*_read)(n_dev* dev, n_u32 addr, n_u32 len, n_u32* rlen, void* data);
	n_code (*_erase)(n_dev* dev, n_u32 addr);
	n_code (*_open)(n_dev* dev);
	n_code (*_close)(n_dev* dev);
};

struct nffsDevice {
	n_dev_param			_param;
	n_dev_vtable		_vtable;

	n_handle			_os_ctx;		/* OS环境 */
	n_handle			_ex_ctx;		/* 额外环境，可以在_vtable->_open中被初始化，使用_vtable->_close释放相关资源 */

	n_block*            _blocks;
	n_block*			_next_block;
	n_block*			_gc_block;

	n_u32               _used_size;
	n_u32               _free_size;
	n_u32               _dirty_size;
	n_u32				_wasted_size;
	n_u32				_erasing_size;
	n_u32				_bad_size;
	n_u32               _unchecked_size;

	n_u32				_nr_free_block;
	n_u32				_nr_erasing_block;

	n_u8				_resv_block_write;
	n_u8				_resv_block_deletion;
	n_u8				_resv_block_gctrigger;
	n_u8				_resv_block_gcbad;
	n_u8				_resv_block_gcmerge;
	n_u8				_vdirty_block_gctrigger;

	n_u32 				_nospc_dirty_size;

	n_u32               _next_ino;
	n_u32				_check_ino;

	n_inode**           _inode_caches;   /* inode哈希表 */
	n_u32               _inode_capacity;

	struct list_head	_clean_list;		/* Blocks 100% full of clean data */
	struct list_head	_very_dirty_list;	/* Blocks with lots of dirty space */
	struct list_head	_dirty_list;		/* Blocks with some dirty space */
	struct list_head	_erasable_list;		/* Blocks which are completely dirty, and need erasing */
	struct list_head	_erasable_pending_wbuf_list;	/* Blocks which need erasing but only after the current wbuf is flushed */
	struct list_head	_erasing_list;		/* Blocks which are currently erasing */
	struct list_head	_erase_checking_list;	/* Blocks which are being checked and marked */
	struct list_head	_erase_pending_list;	/* Blocks which need erasing now */
	struct list_head	_erase_complete_list;	/* Blocks which are erased and need the clean marker written to them */
	struct list_head	_free_list;		/* Blocks which are free and ready to be used */
	struct list_head	_bad_list;		/* Bad blocks. */
	struct list_head	_bad_used_list;		/* Bad blocks with valid data in. */
};

static inline n_code nffsFlashWrite(n_dev* dev, n_u32 off, n_u32 len, n_u32* rlen, void const* buf) {
	return dev->_vtable._write(dev, off, len, rlen, buf);
}

static inline n_code nffsFlashRead(n_dev* dev, n_u32 off, n_u32 len, n_u32* rlen, void* buf) {
	return dev->_vtable._read(dev, off, len, rlen, buf);
}

/* build.c */
extern n_code nffsMount(n_dev** pdev, n_dev_param const* param, n_dev_vtable const* vtable, n_handle os_ctx);
extern void nffsUnmount(n_dev* dev);

/* scan.c */
extern n_code nffsScan(n_dev* dev, n_block* block, n_u8 const* dat);

/* read.c */
extern n_code nffsReadPage(n_file* fp, n_u32 off, void* dst, n_u32 len);

#endif //NFFS_H_
