/**
 * @file nffs_block.h
 * @author Yunhan Xing
 * @version 0.1.0
 * @date 2021/5/10
 * @brief NFFS节点结构体和相关操作定义。
 *
 */

#ifndef NFFS_NODE_H_
#define NFFS_NODE_H_

#include "nffs.h"

#define NFFS_MAGIC		0x1234

enum {
	NODE_TYPE_INODE        = 0xFC,
	NODE_TYPE_STUB         = 0xFD,
	NODE_TYPE_CLEAN_MARKER = 0xFE,
	NODE_TYPE_EMPTY        = 0xFF
};

enum {
	/* all bytes are FF */
	BLOCK_STATE_EMPTY,
	BLOCK_STATE_CLEAN,
	BLOCK_STATE_DIRTY,
};

typedef struct nffsRawNodeAny    n_raw_node_any;
typedef struct nffsRawINode      n_raw_inode;
typedef struct nffsRawDirentNode n_raw_dirent;

#define NFFS_PAGE_SIZE (4 * 1024)
#define NFFS_REF_PER_BLOCK (NFFS_PAGE_SIZE / sizeof(n_raw_node_ref) - 1)
#define NFFS_MIN_DATA_LEN 128

#define ISDIRTY(size) \
    ((size) > sizeof(n_raw_inode) + NFFS_MIN_DATA_LEN)

/**
 *  NFFS闪存形式节点头。
 */
#define NFFS_RAW_NODE_ANY_HEAD \
	n_u16				_magic;			/* 节点魔数，应当是NFFS_MAGIC */								\
	n_u16               _variant;		/* 节点变种标识 */											\
	n_u32               _length;		/* 节点总长度 */												\
	n_u32               _head_crc		/* CRC校验码 */

struct nffsRawNodeAny {
	NFFS_RAW_NODE_ANY_HEAD;
};

struct nffsRawINode {
    NFFS_RAW_NODE_ANY_HEAD;
    n_u32				_ino;            /* 节点序号 */
    n_u32				_cno;            /* 块序号 */
    n_u32				_offset;        /* 该节点存储数据的偏移量 */
    n_u32				_data_size;     /* 该节点数据的大小 */
    n_u32				_node_crc;      /* 整个节点的校验码 */
    n_u8				_data[0];
};

struct nffsRawDirentNode {
	NFFS_RAW_NODE_ANY_HEAD;
	n_u32				_ino;
	n_u32				_pino;
	n_u8				_name_size;     /* 名字的长度 */
	n_u8				_type;
	n_u32				_node_crc;      /* 节点的CRC校验码 */
	n_u32				_name_crc;      /* 目录名的CRC校验码 */

	n_u8				_name[];
};

#define LINK_REF_NODE ((n_i32)-1)
#define EMPTY_REF_NODE ((n_i32)-2)

/**
 * 闪存节点引用。
 */
struct nffsRawNodeRef {
	n_i32				_off;			/* 偏移量和节点状态 */
	n_raw_node_ref*		_next;
};

struct nffsINode {
    n_raw_node_ref*		_ref; /* 当前节点的闪存节点引用，由一个数组链表构成。 */
    n_dirent*			_scan_dents;
    n_inode*			_next;
    n_u32				_ino;
};

struct nffsDirent {
    union {
        n_raw_node_ref*	_ref;
        n_inode*		_in;
    };
    n_dirent*			_next;
    n_u32				_ino;
    n_u32				_name_hash;
    n_u8				_type;
    n_u8				_name[];
};

#define REF_UNCHECKED	0x0
#define REF_PRISTINE	0x1
#define REF_OBSOLETE	0x2
#define REF_NORMAL		0x3

#define REF_EMPTY_NODE  (-2)
#define REF_LINK_NODE   (-1)

#define dirent_node_state(rd)   ((ne2u32((rd)->_ino) ? REF_PRISTINE : REF_NORMAL))

#define is_empty_ref(n)     ((n)->_off == REF_EMPTY_NODE)
#define is_link_ref(n)      ((n)->_off == REF_LINK_NODE)

#define ref_next_raw(n)     (is_link_ref(n) ? (n)->_next : (n) + 1)

#define ref_offset(n)		((n)->_off & ~0x3)
#define ref_status(n)		((n)->_off & 0x3)
#define ref_obsolete(n)		(ref_status(n) == REF_OBSOLETE)
#define mark_ref_normal(n)	((void) ((n)->flash_offset = ref_offset(ref) | REF_NORMAL))

#define ref_totlen(d,b,n)	nffsRefLen(d, b, n)

static inline n_raw_node_ref* ref_next(n_raw_node_ref* ref) {
	ref++;

	if (is_link_ref(ref)) {
		ref = ref->_next;
		if (ref == NULL)
			return NULL;
	}

	return is_empty_ref(ref) ? NULL : ref;
}

#define write_off(d)		((d)->_next_block->_off + (d)->_param._bytes_per_block - (d)->_next_block->_free_size)

struct nffsBlock {
    struct list_head	_list;
    n_raw_node_ref*		_head_node;
    n_raw_node_ref*		_tail_node;
    n_raw_node_ref*		_gc_node;

	n_u32               _offset;		/* offset in flash */

    n_u32				_unchecked_size;
    n_u32				_used_size;
    n_u32				_wasted_size;
    n_u32				_free_size;
    n_u32				_dirty_size;
    n_u32				_allocated_size;
};

#define PAD(x) (((x) + 3) & ~3)

/* nodelist.c */
extern n_code nffsJoinDirtySpace(n_dev* dev, n_block* block, n_u32 size);
extern n_raw_node_ref* nffsJoinNodeRef(n_dev* dev, n_block* block, n_u32 off, n_u32 size, n_inode* node);
extern void nffsFreeAllRawNodeRefs(n_dev* dev);
extern void nffsFreeAllINodes(n_dev* dev);
extern n_inode* nffsGetINode(n_dev* dev, n_u32 ino);
extern void nffsAddINode(n_dev* dev, n_inode* node);
extern void nffsDelINode(n_dev* dev, n_inode* node);
extern void nffsJoinFdList(n_dev* dev, n_dirent* new, n_dirent** list);
extern n_u32 nffsRefLen(n_dev* dev, n_block* block, n_raw_node_ref* ref);

/* alloc.c */
extern n_code nffsPreallocRawNodeRefs(n_dev* dev, n_block* block, n_u32 nr);
extern void nffsFreeRawRefBlock(n_raw_node_ref* block);
extern n_inode* nffsAllocINode(void);
extern void nffsFreeINode(n_inode* node);
extern n_dirent* nffsAllocFullDirent(n_u32 namesize);
extern void nffsFreeFullDirent(n_dirent* dirent);

/* scan.c */
extern n_code nffsScanInodeNode(n_dev *dev, n_block *block, n_raw_inode* ri, n_u32 off);
extern n_code nffsScanDirentNode(n_dev *dev, n_block *block, n_raw_dirent* rd, n_u32 off);

#endif //NFFS_NODE_H_
