/*
 * nodelist.c
 *
 *  Created on: 2021/5/10
 *      Author: ueyudiud
 */

#include "nffs.h"
#include "nffs_node.h"

n_code nffsJoinDirtySpace(n_dev* dev, n_block* block, n_u32 size) {
	if (size == 0)
		return NMF_SUCC;
	nmfAssume(size > block->_free_size, "Dirty space size 0x%x larger than free size 0x%x.",
			size,
			block->_free_size);
	if (block->_tail_node != NULL && ref_obsolete(block->_tail_node)) {
		dev->_dirty_size += size;
		dev->_free_size -= size;
		block->_dirty_size += size;
		block->_free_size -= size;
	}
	else {
		n_u32 off = block->_offset + dev->_param._bytes_per_block - block->_free_size;
		off |= REF_OBSOLETE;
		nffsJoinNodeRef(dev, block, off, size, NULL);
	}
	return NMF_SUCC;
}

n_raw_node_ref* nffsJoinNodeRef(n_dev* dev, n_block* block, n_u32 off, n_u32 size, n_inode* node) {
	n_raw_node_ref* ref = block->_tail_node;

	while (!is_empty_ref(ref)) {
		ref = ref_next_raw(ref);
	}

	ref->_off = off;

	if (block->_head_node == NULL) {
		block->_head_node = ref;
		nmfAssume(block->_offset == ref->_off, "Block offset not match to ref offset\n");
	}
	nmfAssume(ref_offset(ref) == block->_offset + dev->_param._bytes_per_block - block->_free_size,
		   "Adding new ref %p at (0x%08x-0x%08x) not immediately after previous.\n",
		   ref,
		   ref_offset(ref),
		   ref_offset(ref) + size);
	block->_tail_node = ref;

	ref = block->_tail_node;

    nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_INFO, "Last ref node at %p is (%08x,%p)\n",
		   ref, ref->_off, ref->_next);

    while (!is_empty_ref(ref)) {
    	ref = ref_next_raw(ref);
    }

    nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_INFO, "New ref node is %p (%08x becomes %08x,%p) with length of 0x%x\n",
		   ref, ref->_off, off, ref->_next, size);

    ref->_off = off;

    if (!block->_head_node) {
        block->_head_node = ref;
    }

    block->_tail_node = ref;

	if (node != NULL) {
		/* 将节点加入到node的节点链表中 */
		ref->_next = node->_ref;
		node->_ref = ref;
	}
	else {
		ref->_next = NULL;
	}
	/* flags? */

	switch (ref_status(ref)) {
		case REF_UNCHECKED:
			dev->_unchecked_size += size;
			block->_unchecked_size += size;
			break;
		case REF_PRISTINE:
		case REF_NORMAL:
			dev->_used_size += size;
			block->_used_size += size;
			break;
		case REF_OBSOLETE:
			dev->_dirty_size += size;
			block->_dirty_size += size;
			break;
	}
	dev->_free_size -= size;
	block->_free_size -= size;

	return ref;
}

void nffsFreeAllRawNodeRefs(n_dev* dev) {
	for (n_u32 i = 0; i < dev->_param._block_per_total; ++i) {
		n_raw_node_ref* self = dev->_blocks[i]._head_node;
		while (is_link_ref(&self[NFFS_REF_PER_BLOCK])) {
			n_raw_node_ref* next = self->_next;
			nffsFreeRawRefBlock(self);
			self = next;
		}
		dev->_blocks[i]._head_node = NULL;
		dev->_blocks[i]._tail_node = NULL;
	}
}

void nffsFreeAllINodes(n_dev* dev) {
	for (n_u32 i = 0; i < dev->_inode_capacity; ++i) {
		n_inode* self = dev->_inode_caches[i];
		while (self != NULL) {
			n_inode* next = self->_next;
			nffsFreeINode(self);
			self = next;
		}
		dev->_inode_caches[i] = NULL;
	}
}

n_inode* nffsGetINode(n_dev* dev, n_u32 ino) {
	n_inode* node = dev->_inode_caches[ino % dev->_inode_capacity];
	while (node != NULL && node->_ino < ino) {
		node = node->_next;
	}

	return node != NULL && node->_ino == ino ? node : NULL;
}

void nffsAddINode(n_dev* dev, n_inode* node) {
	if (node->_ino == 0) {
		node->_ino = ++dev->_next_ino;
	}

	n_inode** slot = &dev->_inode_caches[node->_ino % dev->_inode_capacity];
	n_inode* prev;

	while ((prev = *slot) != NULL && prev->_ino < node->_ino) {
		slot = &prev->_next;
	}

	node->_next = *slot;
	*slot = node;
}

void nffsDelINode(n_dev* dev, n_inode* node) {
	n_inode** slot = &dev->_inode_caches[node->_ino % dev->_inode_capacity];
	n_inode* prev;

	while ((prev = *slot) != NULL && prev->_ino < node->_ino) {
		slot = &prev->_next;
	}

	if (prev == node) {
		*slot = node->_next;
	}

	/* TODO 应当在某些状态不在这个函数中释放内存 */
	nffsFreeINode(node);
}

void nffsJoinFdList(n_dev* dev, n_dirent* new, n_dirent** list) {
    //TODO finish this function
}

n_u32 nffsRefLen(n_dev* dev, n_block* block, n_raw_node_ref* ref) {
	uint32_t ref_end;
	n_raw_node_ref*next_ref = ref_next(ref);

	if (next_ref)
		ref_end = ref_offset(next_ref);
	else {
		if (block == NULL)
			block = &dev->_blocks[ref->_off / dev->_param._bytes_per_block];

		/* Last node in block. Use free_space */
		nmfAssume(ref != block->_tail_node, "ref %p @0x%08x is not jeb->last_node (%p @0x%08x)\n",
				  ref, ref_offset(ref), block->_tail_node,
				  block->_tail_node ?
				  ref_offset(block->_tail_node) : 0);
		ref_end = block->_offset + dev->_param._bytes_per_block - block->_free_size;
	}
	return ref_end - ref_offset(ref);
}