/*
 * scan.c
 *
 *  Created on: 2021/5/10
 *      Author: ueyudiud
 */

#include "nffs.h"
#include "nffs_node.h"

n_i32 nffsScanCheckBlock(n_dev *dev, n_block *block) {
    /* 需要改进 */
    if (!ISDIRTY(dev->_param._bytes_per_block - (block->_used_size + block->_unchecked_size))) {
        dev->_dirty_size -= block->_dirty_size;
        dev->_wasted_size += dev->_dirty_size;
        block->_wasted_size += block->_dirty_size;
        block->_dirty_size = 0;
        return BLOCK_STATE_CLEAN;
    } else
        return BLOCK_STATE_DIRTY;
}

n_code nffsScan(n_dev *dev, n_block *block, n_u8 const *dat) {
    n_raw_node_any crcnode;

    n_u32 off = 0, base_off = block->_offset;
    n_u32 const blk_len = dev->_param._bytes_per_block;
    /* 当闪存没有被写过时，里面的数据应该全部为0xFF，NFFS以32位为单位判断闪存是否是无效数据 */
    while (off < blk_len && (*(n_u32 *) &dat[off]) == (n_u32) 0xFFFFFFFF)
        off += sizeof(n_u32);

    if (off == blk_len) { /* 当所有数据为FF是，认为该盘块为空 */
        return BLOCK_STATE_EMPTY;
    } else if (off != 0) { /* 否则，如果存在一部分FF，将其标记位脏数据 */
        nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_INFO,
               "Free space at %08x-%08x\n",
               block->_offset,
               block->_offset + off);
        nmfProtect(nffsPreallocRawNodeRefs(dev, block, 1));
        nmfProtect(nffsJoinDirtySpace(dev, block, off));
    }

    /* 转换成整个块的偏移量 */
    off += block->_offset;

    n_u32 prev_off = block->_offset - 1;
    while (off < block->_offset + dev->_param._bytes_per_block) {
        nmfProtect(nffsPreallocRawNodeRefs(dev, block, 2));

        if ((off & 0x3) != 0) { /* 检查偏移量是否是32位对齐 */
            nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_ERROR, "%s(): Offset not aligned.",
                   __func__);
            off = PAD(off); /* 对齐到下一个32位地址 */
            continue;
        } else if (off == prev_off) { /* 判断偏移量是否没有往前移动 */
            nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN, "Data at %08x already been seen, skipping.\n",
                   off);
            nmfProtect(nffsJoinDirtySpace(dev, block, 4));
            off += 4;
            continue;
        }
        prev_off = off;

        if (block->_offset + dev->_param._bytes_per_block < off + sizeof(n_raw_node_any)) { /* 检查是否包含一个节点头 */
            nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_INFO,
                   "Fewer than %zd bytes left to the end of a block. (%x+%x<%x+%zx). Skipped.",
                   sizeof(n_raw_node_any), block->_offset, blk_len, off, sizeof(n_raw_node_any));
            n_u32 s = block->_offset + dev->_param._bytes_per_block - off;
            nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN,
                   "Remaining %u bytes space is not enough for a node header, stop reading.\n",
                   s);
            nmfProtect(nffsJoinDirtySpace(dev, block, s));
            break;
        }

        n_raw_node_any *node = (n_raw_node_any *) &dat[off - base_off];
        if (*(n_u32 *) (&dat[off - base_off]) == 0xFFFFFFFF) {
            // TODO 添加这部分的处理
        }

        if (ne2u16(node->_magic) != NFFS_MAGIC) { /* 检查节点魔数 */
            nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN, "%s(): Bad magic number, got: %04x, expected: %04x.",
                   __func__,
                   ne2u16(node->_magic),
                   NFFS_MAGIC);
            nmfProtect(nffsJoinDirtySpace(dev, block, 4));
            off += 4;
            continue;
        }

        crcnode._magic = node->_magic;
        crcnode._variant = node->_variant;
        crcnode._length = node->_length;
        n_u32 head_crc = nliCrc32(&crcnode, sizeof(n_raw_node_any) - 4);

        if (head_crc != ne2u32(node->_head_crc)) {
            nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN,
                   "%s():Node at 0x%08x {0x%04x, 0x%04x, 0x%08x) has invalid CRC 0x%08x (calculated 0x%08x)",
                   __func__,
                   off,
                   ne2u16(node->_magic),
                   ne2u16(node->_variant),
                   ne2u32(node->_length),
                   ne2u32(node->_head_crc),
                   head_crc);
            nmfProtect(nffsJoinDirtySpace(dev, block, 4));
            off += 4;
            continue;
        }
        if (off + ne2u32(node->_length) >
            block->_offset + dev->_param._bytes_per_block) { /* 检查本块的剩余空间是否能容纳节点 */
            nmfProtect(nffsJoinDirtySpace(dev, block, 4));
            off += 4;
            continue;
        }

        switch (ne2u16(node->_variant)) {
            case NODE_TYPE_INODE:
                nmfProtect(nffsScanInodeNode(dev, block, (void *) node, off));
                off += PAD(ne2u32(node->_length));
                break;
                /* TODO 根据节点类型处理节点 */
        }
    }

    nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_INFO,
           "Block at 0x%08x: free 0x%08x, dirty 0x%08x, unchecked 0x%08x, used 0x%08x, wasted 0x%08x\n",
           block->_offset, block->_free_size, block->_dirty_size,
           block->_unchecked_size, block->_used_size, block->_wasted_size);

    /* 这部分应该暂时没有用 */
    if (block->_wasted_size) {
        block->_dirty_size += block->_wasted_size;
        dev->_dirty_size += block->_wasted_size;
        dev->_wasted_size += block->_wasted_size;
        block->_wasted_size = 0;
    }

    return nffsScanCheckBlock(dev, block);
}

n_inode *nffsScanMakeInoCache(n_dev *dev, n_u32 ino) {
    n_inode *in = nffsGetINode(dev, ino);
    if (in) return in;

    if (ino > dev->_next_ino)
        dev->_next_ino = ino;

    in = nffsAllocINode();
    if (!in) {
        nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_WARN, "%s(): allocation of inode failed\n", __func__);
        return NULL;
    }
    memset(in, 0, sizeof(*in));

    in->_ino = ino;
    in->_ref = (void *) in;
    nffsAddINode(dev, in);
//    if(ino == 1) in->_pino_nlink = 1;
    return in;
}

n_code nffsScanInodeNode(n_dev *dev, n_block *block, n_raw_inode *ri, n_u32 off) {
    n_inode *in;
    n_u32 crc, ino = ne2u32(ri->_ino);

    crc = nliCrc32(ri, sizeof(n_raw_inode) - 8);
    if (crc != ne2u32(ri->_node_crc)) {
        nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_ERROR,
               "%s(): Node CRC failed on node at 0x%08x: Read 0x%08x, calculated 0x%08x\n",
               __func__, off, ne2u32(ri->_node_crc), crc);
        nmfProtect(nffsJoinDirtySpace(dev, block, PAD(ne2u32(ri->_length))));
    }

    if (!(in = nffsGetINode(dev, ino))) {
        in = nffsScanMakeInoCache(dev, ino);
        // TODO 添加分配失败时的处理
    }

    nffsJoinNodeRef(dev, block, off | REF_UNCHECKED, PAD(ne2u32(ri->_length)), in);

    nmfLog(NMF_ENTITY_SELF, NMF_DBG_LEVEL_INFO, "Node is ino #%u, Range 0x%x-0x%x\n",
           ne2u32(ri->_ino), ne2u32(ri->_offset), ne2u32(ri->_offset + ri->_data_size));
    return NMF_SUCC;
}

#define full_name_hash(a, b) (0)

n_u8 nffsDirentNameSize(const n_u8 *name, n_u8 max_size) {
    n_u8 checked_size = 0;
    while (checked_size < max_size) {
        if (name[checked_size] == 0) break;
        checked_size++;
    }
    return checked_size;
}

n_code nffsScanDirentNode(n_dev *dev, n_block *block, n_raw_dirent *rd, n_u32 off) {
    n_u32 crc;
    n_dirent *nd;

    /* 检查节点CRC */
    crc = nliCrc32(rd, sizeof(n_raw_dirent) - 8);
    if (crc != ne2u32(rd->_node_crc)) {
        nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN,
               "%s(): Node CRC failed on node at 0x%08x: Read 0x%08x, calculated 0x%08x\n",
               __func__, off, ne2u32(rd->_node_crc), crc);
        nmfProtect(nffsJoinDirtySpace(dev, block, PAD(ne2u32(rd->_length))));
    }

    /* 正常情况应该有size = rd->_name_size */
    n_u8 size = nffsDirentNameSize(rd->_name, rd->_name_size);
    nd = nffsAllocFullDirent(size + 1);
    memcpy(nd->_name, rd->_name, size);
    nd->_name[size] = 0;

    /* 检查名字CRC */
    crc = nliCrc32(nd->_name, rd->_name_size);
    if (crc != ne2u32(rd->_name_crc)) {
        nmfLog(NMF_ENTITY_DEV, NMF_DBG_LEVEL_WARN,
               "%s(): Name CRC failed on node at 0x%08x: Read 0x%08x, calculated 0x%08x\n"
               "Name for which CRC failed is '%s', ino#%d\n",
               __func__, off, ne2u32(rd->_name_crc), crc, nd->_name, ne2u32(rd->_ino));
        nffsFreeFullDirent(nd);
        nmfProtect(nffsJoinDirtySpace(dev, block, PAD(ne2u32(rd->_length))));
    }

    n_inode *in = makeInoCache(dev, ne2u32(rd->_pino));

    nd->_ref = nffsJoinNodeRef(dev, block, off | dirent_node_state(nd), PAD(ne2u32(rd->_length)), in);

    nd->_next = NULL;
    nd->_ino = ne2u32(rd->_ino);
    nd->_name_hash = full_name_hash(nd->_name, size);
    nd->_type = rd->_type;
    nffsJoinFdList(dev, nd, &in->_scan_dents);
    return NMF_SUCC;
}